"""Testing AnnotatedDataset dataset with eager backend."""
import os
import unittest

import numpy as np

from pytoda.datasets import PolymerTokenizerDataset
from pytoda.tests.utils import TestFileContent


class TestPolymerTokenizerDataset(unittest.TestCase):
    """Testing annotated dataset."""

    def test___len__(self) -> None:

        content_monomer = os.linesep.join(
            ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
        )
        content_catalyst = os.linesep.join(
            [
                'N#CCCC1CCCC1=NNc1ccc([N+](=O)[O-])cc1	CHEMBL543',
                'CC	CHEMBL17',
                'NCCSCCCCC	CHEMBL6402',
            ]
        )

        annotated_content = os.linesep.join(
            [
                'label_0,label_1,monomer,catalyst',
                '2.3,3.4,CHEMBL545,CHEMBL17',
                '4.5,5.6,CHEMBL17564,CHEMBL6402',
                '6.7,7.8,CHEMBL602,CHEMBL6402',
                '6.7,7.8,CHEMBL54556,CHEMBL5434',
            ]
        )

        # longest = 51

        with TestFileContent(content_monomer) as a_test_file:
            with TestFileContent(content_catalyst) as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:
                    polymer_dataset = PolymerTokenizerDataset(
                        *[a_test_file.filename, another_test_file.filename],
                        annotations_filepath=annotation_file.filename,
                        entity_names=['monomer', 'cATalysT'],
                    )

                    self.assertEqual(len(polymer_dataset), 3)

    def test_smiles_params(self) -> None:

        content_monomer = os.linesep.join(
            ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
        )
        content_catalyst = os.linesep.join(
            [
                'N#CCCC1CCCC1=NNc1ccc([N+](=O)[O-])cc1	CHEMBL543',
                'CC	CHEMBL17',
                'NCCSCCCCC	CHEMBL6402',
            ]
        )

        annotated_content = os.linesep.join(
            [
                'label_0,label_1,monomer,catalyst',
                '2.3,3.4,CHEMBL545,CHEMBL17',
                '4.5,5.6,CHEMBL17564,CHEMBL6402',
                '6.7,7.8,CHEMBL602,CHEMBL6402',
            ]
        )

        with TestFileContent(content_monomer) as a_test_file:
            with TestFileContent(content_catalyst) as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:
                    polymer_dataset = PolymerTokenizerDataset(
                        *[a_test_file.filename, another_test_file.filename],
                        annotations_filepath=annotation_file.filename,
                        entity_names=['monomer', 'cATalysT'],
                        all_bonds_explicit=True,
                        all_hs_explicit=[True, False],
                        sanitize=[True, False],
                        padding_length=[9, None],
                    )

                    pad_ind = polymer_dataset.smiles_language.padding_index
                    monomer_start_ind = polymer_dataset.smiles_language.token_to_index[
                        '<MONOMER_START>'
                    ]
                    monomer_stop_ind = polymer_dataset.smiles_language.token_to_index[
                        '<MONOMER_STOP>'
                    ]
                    catalyst_start_ind = polymer_dataset.smiles_language.token_to_index[
                        '<CATALYST_START>'
                    ]
                    catalyst_stop_ind = polymer_dataset.smiles_language.token_to_index[
                        '<CATALYST_STOP>'
                    ]
                    ch3_ind = polymer_dataset.smiles_language.token_to_index['[CH3]']
                    oh_ind = polymer_dataset.smiles_language.token_to_index['[OH]']
                    ch2_ind = polymer_dataset.smiles_language.token_to_index['[CH2]']
                    b_ind = polymer_dataset.smiles_language.token_to_index['-']
                    c_ind = polymer_dataset.smiles_language.token_to_index['C']

                    # test first sample
                    monomer, catalyst, labels = polymer_dataset[0]

                    # CCO -> CH3CH2OH
                    self.assertEqual(
                        monomer.numpy().flatten().tolist(),
                        [
                            pad_ind,
                            pad_ind,
                            monomer_start_ind,
                            ch3_ind,
                            b_ind,
                            ch2_ind,
                            b_ind,
                            oh_ind,
                            monomer_stop_ind,
                        ],
                    )
                    self.assertEqual(
                        polymer_dataset.smiles_language.token_indexes_to_smiles(
                            monomer
                        ),
                        '[CH3]-[CH2]-[OH]',
                    )
                    # CC
                    self.assertEqual(
                        catalyst.numpy().flatten().tolist(),
                        [
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            catalyst_start_ind,
                            c_ind,
                            b_ind,
                            c_ind,
                            catalyst_stop_ind,
                        ],
                    )
                    self.assertEqual(
                        polymer_dataset.smiles_language.token_indexes_to_smiles(
                            catalyst
                        ),
                        'C-C',
                    )
                    self.assertTrue(
                        np.allclose(labels.numpy().flatten().tolist(), [2.3, 3.4])
                    )

    def test___getitem__(self) -> None:
        """Test __getitem__."""
        content_monomer = os.linesep.join(
            ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
        )
        content_catalyst = os.linesep.join(
            ['c1ccc([N+])cc1	CHEMBL543', 'CC	CHEMBL17', 'NCCSCCCCC	CHEMBL6402']
        )

        annotated_content = os.linesep.join(
            [
                'label_0,label_1,monomer,catalyst',
                '2.3,3.4,CHEMBL545,CHEMBL17',
                '4.5,5.6,CHEMBL17564,CHEMBL6402',
                '6.7,7.8,CHEMBL602,CHEMBL6402',
            ]
        )
        with TestFileContent(content_monomer) as a_test_file:
            with TestFileContent(content_catalyst) as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:
                    polymer_dataset = PolymerTokenizerDataset(
                        *[a_test_file.filename, another_test_file.filename],
                        annotations_filepath=annotation_file.filename,
                        entity_names=['monomer', 'cATalysT'],
                        remove_bonddir=True,
                    )

                    pad_ind = polymer_dataset.smiles_language.padding_index
                    monomer_start_ind = polymer_dataset.smiles_language.token_to_index[
                        '<MONOMER_START>'
                    ]
                    monomer_stop_ind = polymer_dataset.smiles_language.token_to_index[
                        '<MONOMER_STOP>'
                    ]
                    catalyst_start_ind = polymer_dataset.smiles_language.token_to_index[
                        '<CATALYST_START>'
                    ]
                    catalyst_stop_ind = polymer_dataset.smiles_language.token_to_index[
                        '<CATALYST_STOP>'
                    ]
                    c_ind = polymer_dataset.smiles_language.token_to_index['C']
                    o_ind = polymer_dataset.smiles_language.token_to_index['O']
                    n_ind = polymer_dataset.smiles_language.token_to_index['N']
                    s_ind = polymer_dataset.smiles_language.token_to_index['S']

                    # test first sample
                    monomer, catalyst, labels = polymer_dataset[0]

                    self.assertEqual(
                        monomer.numpy().flatten().tolist(),
                        [pad_ind] * 8
                        + [monomer_start_ind, c_ind, c_ind, o_ind, monomer_stop_ind],
                    )
                    self.assertEqual(
                        catalyst.numpy().flatten().tolist(),
                        [
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            catalyst_start_ind,
                            c_ind,
                            c_ind,
                            catalyst_stop_ind,
                        ],
                    )
                    self.assertTrue(
                        np.allclose(labels.numpy().flatten().tolist(), [2.3, 3.4])
                    )

                    monomer, catalyst, labels = polymer_dataset[2]

                    self.assertEqual(
                        monomer.numpy().flatten().tolist(),
                        [pad_ind] * 7
                        + [
                            monomer_start_ind,
                            n_ind,
                            c_ind,
                            c_ind,
                            s_ind,
                            monomer_stop_ind,
                        ],
                    )
                    self.assertEqual(
                        catalyst.numpy().flatten().tolist(),
                        [
                            pad_ind,
                            pad_ind,
                            catalyst_start_ind,
                            n_ind,
                            c_ind,
                            c_ind,
                            s_ind,
                            c_ind,
                            c_ind,
                            c_ind,
                            c_ind,
                            c_ind,
                            catalyst_stop_ind,
                        ],
                    )
                    self.assertTrue(
                        np.allclose(labels.numpy().flatten().tolist(), [6.7, 7.8])
                    )

    def test___getitem___with_annotation_column_names(self) -> None:
        """Test __getitem__ with annotations_column_names in the annotation."""
        content_monomer = os.linesep.join(
            ['CCO	CHEMBL545', 'C	CHEMBL17564', 'CO	CHEMBL14688', 'NCCS	CHEMBL602']
        )
        content_catalyst = os.linesep.join(
            ['c1ccc([N+])cc1	CHEMBL543', 'CC	CHEMBL17', 'NCCSCCCCC	CHEMBL6402']
        )
        annotated_content = os.linesep.join(
            [
                'index,label_0,label_1,monomer,catalyst',
                '0,2.3,3.4,CHEMBL545,CHEMBL6402',
                '1,4.5,5.6,CHEMBL17564,CHEMBL543',
                '1,6.7,7.8,CHEMBL602,CHEMBL17',
            ]
        )
        with TestFileContent(content_monomer) as a_test_file:
            with TestFileContent(content_catalyst) as another_test_file:
                with TestFileContent(annotated_content) as annotation_file:
                    polymer_dataset = PolymerTokenizerDataset(
                        *[a_test_file.filename, another_test_file.filename],
                        annotations_filepath=annotation_file.filename,
                        entity_names=['monomer', 'cATalysT'],
                        annotations_column_names=['label_0'],
                        remove_bonddir=True,
                    )
                    pad_ind = polymer_dataset.smiles_language.padding_index
                    monomer_start_ind = polymer_dataset.smiles_language.token_to_index[
                        '<MONOMER_START>'
                    ]
                    monomer_stop_ind = polymer_dataset.smiles_language.token_to_index[
                        '<MONOMER_STOP>'
                    ]
                    catalyst_start_ind = polymer_dataset.smiles_language.token_to_index[
                        '<CATALYST_START>'
                    ]
                    catalyst_stop_ind = polymer_dataset.smiles_language.token_to_index[
                        '<CATALYST_STOP>'
                    ]
                    c_ind = polymer_dataset.smiles_language.token_to_index['C']
                    o_ind = polymer_dataset.smiles_language.token_to_index['O']
                    n_ind = polymer_dataset.smiles_language.token_to_index['N']
                    s_ind = polymer_dataset.smiles_language.token_to_index['S']

                    # test first sample
                    monomer, catalyst, labels = polymer_dataset[0]

                    self.assertEqual(
                        monomer.numpy().flatten().tolist(),
                        [pad_ind] * 8
                        + [monomer_start_ind, c_ind, c_ind, o_ind, monomer_stop_ind],
                    )
                    self.assertEqual(
                        catalyst.numpy().flatten().tolist(),
                        [
                            pad_ind,
                            pad_ind,
                            catalyst_start_ind,
                            n_ind,
                            c_ind,
                            c_ind,
                            s_ind,
                            c_ind,
                            c_ind,
                            c_ind,
                            c_ind,
                            c_ind,
                            catalyst_stop_ind,
                        ],
                    )
                    self.assertTrue(
                        np.allclose(labels.numpy().flatten().tolist(), [2.3])
                    )

                    monomer, catalyst, labels = polymer_dataset[2]

                    self.assertEqual(
                        monomer.numpy().flatten().tolist(),
                        [pad_ind] * 7
                        + [
                            monomer_start_ind,
                            n_ind,
                            c_ind,
                            c_ind,
                            s_ind,
                            monomer_stop_ind,
                        ],
                    )
                    self.assertEqual(
                        catalyst.numpy().flatten().tolist(),
                        [
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            pad_ind,
                            catalyst_start_ind,
                            c_ind,
                            c_ind,
                            catalyst_stop_ind,
                        ],
                    )
                    self.assertTrue(
                        np.allclose(labels.numpy().flatten().tolist(), [6.7])
                    )


if __name__ == '__main__':
    unittest.main()



"""Implementation of base classes working with datasets."""
import bisect

import pandas as pd
from torch.utils.data import ConcatDataset, Dataset

from ..types import Any, Hashable, Iterator, List, Tuple, Union


class KeyDataset(Dataset):
    """
    Base Class for Datsets with both integer index and item identifier key.

    Implicit abstract methods are:
    `__len__(self)` https://github.com/pytorch/pytorch/blob/66a20c259b3b2063e59102ab23f3fb34fc819455/torch/utils/data/sampler.py#L23
    `__getitem__(self, index: int)` is inherited

    Default implementations to index using key and getting all keys are
    provided but should be overloaded when possible as calls to `get_item`
    and `get_key` might be expensive.

    The keys are expected to be unique. Call `has_duplicate_keys` to make sure.
    If there are duplicate keys, on lookup generally the first one found will
    be used, but there are no guarantees.
    """

    def __add__(self, other):
        return ConcatKeyDataset([self, other])

    def get_key(self, index: int) -> Hashable:
        """Get key from integer index."""
        raise NotImplementedError

    def get_index(self, key: Hashable) -> int:
        """Get index for first datum mapping to the given key."""
        raise NotImplementedError

    def get_item_from_key(self, key: Hashable) -> Any:
        """Get item via key"""
        return self.__getitem__(self.get_index(key))

    def keys(self) -> Iterator:
        """Default iterator of keys by iterating over dataset indexes."""
        for index in range(len(self)):
            yield self.get_key(index)

    @property
    def has_duplicate_keys(self) -> bool:
        """Check whether each key is unique."""
        return pd.Index(self.keys()).has_duplicates


class DatasetDelegator(Dataset):
    """
    Base class for KeyDataset attribute accesses from `self.dataset`.

    The attributes/methods to delegate are stored to allow explicit filtering
    and addition to class documentation.

    Source: https://www.fast.ai/2019/08/06/delegation/
    """

    # built-in methods need to be defined explicitly
    # https://stackoverflow.com/a/57589213
    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, index: int):
        return self.dataset[index]

    # returned dataset is not a Delegator
    def __add__(self, other):
        return ConcatKeyDataset([self, other])

    # explicit implementation to ensure use of top level __getitem__
    # (possibly overloaded) vs using datasets.__getitem__
    def get_item_from_key(self, key: Hashable) -> Any:
        """Get datum mapping to the given key."""
        return self.__getitem__(self.get_index(key))

    @staticmethod
    def _delegation_filter(method_name):
        """To remove unwanted attributes/methods from being delegated."

        Args:
            method_name (str): attribute/method of an instance.

        Returns:
            bool: wether to delegate method.
        """
        # do not delegate to all private attributes
        if method_name.startswith('_'):
            return False
        else:
            return True

    # call super().__getattribute__ is needed in place of self, as
    # calls to self could  lead to infinite loops, e.g. with copy

    @property
    def _delegatable(self):
        return [
            o
            for o in dir(
                super().__getattribute__('dataset')
            )  # other AttributeErrors are masked here if dataset is not set  # noqa
            if self._delegation_filter(o)
        ]

    # delegation, i.e. in case method not defined in class or class hierarchy
    def __getattr__(self, k):
        if k in super().__getattribute__('_delegatable'):
            return getattr(super().__getattribute__('dataset'), k)
        raise AttributeError(k)

    def __dir__(self):
        return dir(type(self)) + list(self.__dict__.keys()) + self._delegatable


class TransparentConcatDataset(ConcatDataset):
    """Extension of ConcatDataset with transparent indexing."""

    def __init__(self, datasets: List[Dataset]):
        super().__init__(datasets)

    def get_index_pair(self, idx: int) -> Tuple[int, int]:
        """
        Get dataset and sample indexes.

        Args:
            idx (int): index in the concatenated dataset.

        Returns:
            Tuple[int, int]: dataset and sample index.
        """
        # related request https://github.com/pytorch/pytorch/issues/32034
        if idx < 0:
            if -idx > len(self):
                raise ValueError(
                    'absolute value of index should not exceed dataset length'
                )
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        return (dataset_idx, sample_idx)


class ConcatKeyDataset(TransparentConcatDataset, KeyDataset):
    """
    Extension of ConcatDataset with transparent indexing supporting KeyDataset

    The keys are expected to be unique. If there are duplicate keys, on lookup
    the first one found will be used by default.
    """

    def __init__(self, datasets: List['AnyBaseDataset']):
        """
        Initialize the ConcatKeyDataset.

        Args:
            datasets (List[AnyBaseDataset]): a list of datasets.
        """
        super().__init__(datasets)
        # __getitem__ and __len__ implementation from ConcatDataset

    # the following require dataset to implement get_key and get_index
    def get_key_pair(self, index: int) -> Tuple[int, Hashable]:
        """Get dataset index key from integer index."""
        dataset_idx, sample_idx = self.get_index_pair(index)
        return dataset_idx, self.datasets[dataset_idx].get_key(sample_idx)

    def get_key(self, index: int) -> Hashable:
        """Get key from integer index."""
        return self.get_key_pair(index)[1]

    def get_index(self, key: Hashable) -> int:
        """Get index for first datum mapping to the given key."""
        for dataset_idx, dataset in enumerate(self.datasets):
            try:
                sample_idx = dataset.get_index(key)
            except KeyError:
                continue
            break
        else:
            raise KeyError(f'key {key} not found in datasets.')
        # return index from index_pair
        if dataset_idx == 0:
            return sample_idx
        else:
            return sample_idx + self.cumulative_sizes[dataset_idx - 1]

    def get_item_from_key(self, key: Hashable) -> Any:
        """Get datum mapping to the given key."""
        return self.__getitem__(self.get_index(key))

    def keys(self) -> Iterator:
        """Default generator of keys by iterating over dataset."""
        for dataset in self.datasets:
            for key in dataset.keys():
                yield key

    # default has_duplicate_keys via keys()


AnyBaseDataset = Union[KeyDataset, DatasetDelegator, ConcatKeyDataset]



"""Implementation of ProteinSequenceDataset."""
import logging
from typing import Dict, Iterable

import numpy as np
import torch
from importlib_resources import files

from pytoda.warnings import device_warning

from ..proteins.protein_feature_language import ProteinFeatureLanguage
from ..proteins.protein_language import ProteinLanguage
from ..proteins.transforms import (
    MutateResidues,
    ProteinAugmentFlipSubstrs,
    ProteinAugmentSwapSubstrs,
    ReplaceByFullProteinSequence,
    SequenceToTokenIndexes,
)
from ..transforms import (
    AugmentByReversing,
    Compose,
    DiscardLowercase,
    ExtractFromDict,
    LeftPadding,
    ListToTensor,
    Randomize,
    ToTensor,
    ToUpperCase,
)
from ._fasta_eager_dataset import _FastaEagerDataset
from ._fasta_lazy_dataset import _FastaLazyDataset
from ._smi_eager_dataset import _SmiEagerDataset
from ._smi_lazy_dataset import _SmiLazyDataset
from .base_dataset import DatasetDelegator, KeyDataset
from .utils import concatenate_file_based_datasets

logger = logging.getLogger(__name__)

SEQUENCE_DATASET_IMPLEMENTATIONS = {  # get class and acceptable keywords
    '.csv': {
        'eager': (_SmiEagerDataset, {'index_col', 'names'}),
        'lazy': (_SmiLazyDataset, {'chunk_size', 'index_col', 'names'}),
    },  # .smi like .csv
    '.smi': {
        'eager': (_SmiEagerDataset, {'index_col', 'names'}),
        'lazy': (_SmiLazyDataset, {'chunk_size', 'index_col', 'names'}),
    },
    '.fasta': {
        'eager': (_FastaEagerDataset, {'gzipped', 'name'}),
        'lazy': (
            _FastaLazyDataset,
            {
                'name',
                # args to pyfaidx.Fasta
                'default_seq',
                'key_function',
                'as_raw',
                'strict_bounds',
                'read_ahead',
                'mutable',
                'split_char',
                'duplicate_action',
                'filt_function',
                'one_based_attributes',
                'read_long_names',
                'sequence_always_upper',
                'rebuild',
                'build_index',
            },
        ),
    },
    '.fasta.gz': {
        'eager': (_FastaEagerDataset, {'gzipped', 'name'}),
        # requires Biopython installation
        'lazy': (
            _FastaLazyDataset,
            {
                'name',
                # args to pyfaidx.Fasta
                'default_seq',
                'key_function',
                'as_raw',
                'strict_bounds',
                'read_ahead',
                'mutable',
                'split_char',
                'duplicate_action',
                'filt_function',
                'one_based_attributes',
                'read_long_names',
                'sequence_always_upper',
                'rebuild',
                'build_index',
            },
        ),
    },
}


def protein_sequence_dataset(
    *filepaths: str, filetype: str, backend: str, **kwargs
) -> KeyDataset:
    """
    Return a dataset of protein sequences.

    Args:
        filepaths (Iterable[str]): Paths to the files containing the protein sequences.
        filetype (str): The filetype of the protein sequences.
        backend (str): The backend to use for the dataset.
        kwargs: Keyword arguments for the dataset.
    """
    try:
        # hardcoded factory
        dataset_class, valid_keys = SEQUENCE_DATASET_IMPLEMENTATIONS[filetype][backend]
    except KeyError:
        raise ValueError(  # filetype checked already
            f'backend {backend} not supported for {filetype}.'
        )

    kwargs['gzipped'] = True if filetype == '.fasta.gz' else False
    # prune unsupported arguments
    kwargs = dict((k, v) for k, v in kwargs.items() if k in valid_keys)
    kwargs['name'] = 'Sequence'

    return concatenate_file_based_datasets(
        filepaths=filepaths, dataset_class=dataset_class, **kwargs
    )


class ProteinSequenceDataset(DatasetDelegator):
    """Protein Sequence dataset using a Language to transform sequences."""

    def __init__(
        self,
        *filepaths: str,
        filetype: str = '.smi',
        protein_language: ProteinLanguage = None,
        amino_acid_dict: str = 'iupac',
        padding: bool = True,
        padding_length: int = None,
        add_start_and_stop: bool = False,
        augment_by_revert: bool = False,
        sequence_augment: Dict = {},
        protein_keep_only_uppercase: bool = False,
        randomize: bool = False,
        backend: str = 'eager',
        iterate_dataset: bool = False,
        name: str = 'protein-sequences',
        device: torch.device = None,
        **kwargs,
    ) -> None:
        """
        Initialize a Protein Sequence dataset.

        Args:
            *filepaths (Files): paths to .smi, .csv/.fasta/.fasta.gz file
                with the sequences.
            filetype (str): From {.smi, .csv, .fasta, .fasta.gz}.
            protein_language (ProteinLanguage): a ProteinLanguage (or child)
                instance, e.g. ProteinFeatureLanguage. Defaults to None,
                creating a default instance.
            amino_acid_dict (str): Type of dictionary used for amino acid sequences.
                Defaults to 'iupac', alternative is 'unirep' or 'human-kinase-alignment'
            padding (bool): pad sequences to longest in the protein language.
                Defaults to True.
            padding_length (int): manually sets number of applied paddings,
                applies only if padding is True. Defaults to None.
            add_start_and_stop (bool): add start and stop token indexes.
                Defaults to False.
            augment_by_revert (bool): perform Protein augmentation by reverting
                Sequences. Defaults to False.
            randomize (bool): perform a true randomization of Protein tokens.
                Defaults to False.
            sequence_augment (Dict): a dictionary to specify additional sequence
                augmentation. Defaults to {}. Items can be:
                    `alignment_path`: A path (str) to a `.smi` (or `.tsv`) file
                        with alignment information that specifies which residues of the
                        sequence are to be used. E.g., to extract active site sequences
                        from full proteins. Do not use a header in the file.
                        1. column has to be the full protein sequence (use upper
                            case only for residues to be used). E.g., ggABCggDEFgg
                        2. column has to be the condensed sequence (ABCDEF).
                        3. column has to be the protein identifier.
                        NOTE: Such a file is *necessary* to apply all augmentation types
                        specified in this dictionary (`sequence_augment`).
                        NOTE: Unless specified, this defaults to
                        `kinase_activesite_alignment.smi`, a file in
                        `pytoda.proteins.metadata` that can ONLY be used to
                        extract active site sequences of human kinases (based on the
                        active site definition in Sheridan et al. (2009, JCIM) and
                        Martin & Mukherjee (2012, JCIM).
                    `discard_lowercase`: A (bool) specifying whether all lowercase
                        characters (residues) in the sequence should be discarded.
                        NOTE: This defaults to True.
                    `flip_substrings`: A probability (float) to flip each contiguous
                        upper-case substring (e.g., an active site substring that lies
                        contiguously in the original sequence).
                        Defaults to 0.0, i.e., no flipping.
                        E.g., ABCDEF could become CBADEF or CBAFED or ABCFED if the
                        original sequence is ggABCggDEFgg.
                    `swap_substrings`: A probability (float) to swap neighboring
                        substrings. Defaults to 0.0, i.e., no swapping.
                        E.g., ABCDEF could become DEFABC if the original sequence is
                        ggABCggDEFgg.
                    `noise`: A 2-Tuple of (float, float) that specifies the probability
                        for a random, single-residue mutation inside and outside the
                        relevent part. Defaults to (0.0, 0.0), i.e., no noise.
                        E.g., with (0.0, 0.5),  ggABCggDEFgg could become hgABCgbDEFgg.
            iterate_dataset (bool): whether to go through all items in the dataset
                to detect unknown characters, find longest sequence and checks
                passed padding length if applicable. Defaults to False.
            backend (str): memory management backend.
                Defaults to eager, prefer speed over memory consumption.
            name (str): name of the ProteinSequenceDataset.
            device (torch.device): DEPRECATED
            kwargs (dict): additional arguments for dataset constructor.
        """

        device_warning(device)
        # Parse language object and data paths
        self.filepaths = filepaths
        self.filetype = filetype
        self.backend = backend

        assert filetype in [
            '.csv',
            '.smi',
            '.fasta',
            '.fasta.gz',
        ], f'Unknown filetype given {filetype}'
        self.name = name

        # setup dataset
        self._setup_dataset(**kwargs)
        self.cumulative_sizes = np.cumsum([0] + [len(s) for s in self.datasets])

        DatasetDelegator.__init__(self)  # delegate to self.dataset
        if self.has_duplicate_keys:
            raise KeyError(f'Please remove duplicates from your {self.filetype} file.')

        if protein_language is None:
            self.protein_language = ProteinLanguage(
                amino_acid_dict=amino_acid_dict, add_start_and_stop=add_start_and_stop
            )
        else:
            self.protein_language = protein_language
            assert (
                add_start_and_stop == protein_language.add_start_and_stop
            ), f'add_start_and_stop was "{add_start_and_stop}", but given '
            f'Protein Language has {protein_language.add_start_and_stop}.'

        if iterate_dataset or not protein_language:
            tokens = set(self.protein_language.token_to_index.keys())
            for sequence in self.dataset:
                # sets max_token_sequence_length
                self.protein_language.add_sequence(sequence)
                seq_tokens = set(list(sequence))
                if seq_tokens - tokens != set():
                    logger.error(
                        'Found unknown token(s): %s', list(seq_tokens - tokens)
                    )
        elif (
            not iterate_dataset
            and protein_language
            and protein_language.max_token_sequence_length < 3
        ):
            raise ValueError(
                'If provided ProteinLanguage is empty, set iterate_dataset to True'
            )

        # Set up transformation paramater
        self.padding = padding
        self.padding_length = (
            self.protein_language.max_token_sequence_length
            if padding_length is None
            else padding_length
        )
        if self.padding_length < self.protein_language.max_token_sequence_length:
            logger.warning(
                f'WARNING: Passed padding length was {padding_length} but '
                'protein language has padding length: '
                f'{self.protein_language.max_token_sequence_length}. '
                'Some sequences might get truncated.'
            )
        self.randomize = randomize
        self.augment_by_revert = augment_by_revert

        # Sequence-based augmentation
        self.sequence_augment = sequence_augment
        transforms = self.setup_sequence_augmentation(sequence_augment)

        if self.augment_by_revert:
            transforms += [AugmentByReversing()]

        self.language_transforms = Compose(transforms.copy())

        transforms += [SequenceToTokenIndexes(protein_language=self.protein_language)]
        if self.randomize:
            transforms += [Randomize()]
        if self.padding:
            transforms += [
                LeftPadding(
                    padding_length=self.padding_length,
                    padding_index=self.protein_language.token_to_index['<PAD>'],
                )
            ]
        if isinstance(self.protein_language, ProteinFeatureLanguage):
            transforms += [ListToTensor()]
        elif isinstance(self.protein_language, ProteinLanguage):
            transforms += [ToTensor()]
        else:
            raise TypeError(
                'Please choose either ProteinLanguage or ' 'ProteinFeatureLanguage'
            )
        self.transform = Compose(transforms)

    def _setup_dataset(self, **kwargs) -> None:
        """Setup the dataset."""
        self.dataset = protein_sequence_dataset(
            *self.filepaths, filetype=self.filetype, backend=self.backend, **kwargs
        )

    def setup_sequence_augmentation(self, sequence_augment: Dict):
        """
        Setup the sequence augmentation.

        Args:
            sequence_augment: A dictionary to specify the sequence augmentation
                strategies. For details see the constructor docs.
        """
        if not isinstance(sequence_augment, Dict):
            raise TypeError(
                f"Pass sequence_augment as Dict not {type(sequence_augment)}"
            )

        # Build up cascade of Protein transformations
        transforms = []

        # No work to do
        if sequence_augment == {}:
            transforms += [ExtractFromDict(key='sequence')]
            return transforms

        # Start composing transforms
        self.alignment_path = sequence_augment.pop(
            'alignment_path',
            files('pytoda.proteins.metadata').joinpath(
                'kinase_activesite_alignment.smi'
            ),
        )

        transforms += [ReplaceByFullProteinSequence(self.alignment_path)]

        # Flip substrings
        self.flip_substrings = sequence_augment.pop('flip_substrings', 0.0)
        transforms += [ProteinAugmentFlipSubstrs(p=self.flip_substrings)]

        # Swap substrings
        self.swap_substrings = sequence_augment.pop('swap_substrings', 0.0)
        transforms += [ProteinAugmentSwapSubstrs(self.swap_substrings)]

        # Inject noise (single residue mutation)
        self.noise = sequence_augment.pop('noise', (0.0, 0.0))
        if not isinstance(self.noise, Iterable) or len(self.noise) != 2:
            raise TypeError(f'Noise has to be Iterable of length 2 not {self.noise}')
        transforms += [MutateResidues(*self.noise)]

        # Prune lowercase characters
        self.discard_lowercase = sequence_augment.pop('discard_lowercase', True)
        if self.discard_lowercase:
            transforms += [DiscardLowercase()]

        """
        In the future this could be removed to distinguish between upper/lowercase
        characters but it would necessitate changes in protein language.
        """
        # Always convert to uppercase
        transforms += [ToUpperCase()]

        for k, v in sequence_augment.items():
            logger.warning(f"Ignoring sequence_augment item: {k}: {v}.")

        return transforms

    def __getitem__(self, index: int) -> torch.Tensor:
        """
        Generates one sample of data.

        Args:
            index (int): index of the sample to fetch.

        Returns:
            torch.Tensor: a torch tensor of token indexes,
                for the current sample.
        """
        sample_dict = dict(sequence=self.dataset[index], id=self.dataset.get_key(index))
        return self.transform(sample_dict)
